#include <iostream>
#include <string>
using namespace std;

#include "GL/glew.h"

#ifdef __APPLE__
#include <GLUT/glut.h>
#else
#include <GL/glut.h>
#endif

#include "parameters.h"
#include "simple_gl_common.h"
#include "drawtext.h"
#include "simple_gl.h"
#include "ui.h"
#include "resource.h"

#include "main.h"

extern bool init();
extern void draw();
extern void exit();


GLuint g_fontBase;
static int g_width, g_height;

const int IS_DOWN_SIZE = 300;
static bool isDown[IS_DOWN_SIZE];

const int REFRESH_MS = 25;

void setStartupText(const std::string& s)
{
  cout << s << endl;
}

bool isKeyPressed(int k)
{
  if(k >= IS_DOWN_SIZE) return false;
  return isDown[k];
}

void flush()
{
  //draw strings
  if(isKeyPressed(BKEY_TAB))
  {
    renderTexts(g_width, g_height, 7, 12,
             g_fontBase, 1.f/getAverageSecondsPerFrame());
  }

  glutSwapBuffers();
}

void onDisplay()
{
  int start = glutGet(GLUT_ELAPSED_TIME);
  
  draw();
  
  setLastFrameSeconds((glutGet(GLUT_ELAPSED_TIME) - start)/1000.f);
  //setLastFrameSeconds(REFRESH_MS/1000.f);
}

void onReshape(int w, int h)
{
  g_width = w; g_height = h;
  updateProjectionMatrix(w, h);
}

void onMenu(int option)
{
  switch(option)
  {
    case MENU_FILE_EXPORT_SHADERS:
      menuFileExportShaders();
      break;

    case MENU_FILE_REIMPORT_SHADERS:
      menuFileReimportShaders();
      break;

    case MENU_FILE_REGENERATE_SHADERS:
      menuFileRegenerateShaders();
      break;

    case MENU_FILE_EXPORT_MODEL:
      menuFileExportTextures(g_models.back().bmdFileName + "_export.3ds");
      break;

    case MENU_FILE_EXPORT_TEXTURES:
      menuFileExportTextures(g_models.back().bmdFileName + "_tex.tga");
      break;

    case MENU_DEBUG_SECTIONINFO:
      menuDebugSectioninfo();
      break;
  }
}

void setupMenu()
{
  glutCreateMenu(onMenu);
  glutAddMenuEntry("Export Shaders", MENU_FILE_EXPORT_SHADERS);
  glutAddMenuEntry("Re-Import Shaders", MENU_FILE_REIMPORT_SHADERS);
  glutAddMenuEntry("Regenerate Shaders", MENU_FILE_REGENERATE_SHADERS);

  glutAddMenuEntry("Dump model", MENU_FILE_EXPORT_MODEL);
  glutAddMenuEntry("Dump textures", MENU_FILE_EXPORT_TEXTURES);

  glutAddMenuEntry("Print debug stuff", MENU_DEBUG_SECTIONINFO);
  
  glutAttachMenu(GLUT_RIGHT_BUTTON);
}

void init3D()
{
  glewInit();
  
  g_fontBase = glGenLists(256);
  for(int i = 0; i < 256; ++i)
  {
    glNewList(g_fontBase + i, GL_COMPILE);
    glutBitmapCharacter(GLUT_BITMAP_HELVETICA_12, (char)i);
    glEndList();
  }
}

void updateModifiers()
{
  //this is a hack: glut is not able to tell us if ctrl, shift or alt
  //are pressed, they are only sent as modifiers with keyboard events
  //for other keys. so this method can't detect if a modifier is pressed
  //when no other key is pressed at the same time
  int mod = glutGetModifiers();
  isDown[BKEY_SHIFT] = (mod & GLUT_ACTIVE_SHIFT) != 0;
  isDown[BKEY_CONTROL] = (mod & GLUT_ACTIVE_CTRL) != 0;
  isDown[BKEY_ALT] = (mod & GLUT_ACTIVE_ALT) != 0;
}

void onKeyboard(unsigned char key, int x, int y)
{
  updateModifiers();
  isDown[toupper(key)] = true;
}

void onKeyboardUp(unsigned char key, int x, int y)
{
  updateModifiers();
  isDown[toupper(key)] = false;
}

int specialCode(int glutCode)
{
  switch(glutCode)
  {
    case GLUT_KEY_LEFT: return BKEY_LEFT;
    case GLUT_KEY_RIGHT: return BKEY_RIGHT;
    case GLUT_KEY_DOWN: return BKEY_DOWN;
    case GLUT_KEY_UP: return BKEY_UP;
    case GLUT_KEY_PAGE_UP: return BKEY_PG_UP;
    case GLUT_KEY_PAGE_DOWN: return BKEY_PG_DOWN;
    case GLUT_KEY_F1: return BKEY_F1;
    case GLUT_KEY_F2: return BKEY_F2;
    default: return IS_DOWN_SIZE - 1;
  }
}

void onSpecial(int key, int x, int y)
{
  updateModifiers();
  isDown[specialCode(key)] = true;
}

void onSpecialUp(int key, int x, int y)
{
  updateModifiers();
  isDown[specialCode(key)] = false;
}

void onTimer(int val)
{
  handleCamera();
  glutPostRedisplay();
  glutTimerFunc(REFRESH_MS, onTimer, 0);
}

int main(int argc, char* argv[])
{
  glutInit(&argc, argv);
  glutInitDisplayMode(GLUT_DOUBLE | GLUT_DEPTH | GLUT_RGBA);
  glutInitWindowSize(640, 480);

  glutCreateWindow("bmdview2");

  setupMenu();

  init3D();

  parseParameters(argc, argv);
  if(!init())
  {
    cerr << "Argh. Something went wrong" << endl;
    return -1;
  }

  glutDisplayFunc(onDisplay);
  glutReshapeFunc(onReshape);
  glutKeyboardFunc(onKeyboard);
  glutKeyboardUpFunc(onKeyboardUp);
  glutSpecialFunc(onSpecial);
  glutSpecialUpFunc(onSpecialUp);
  
  glutTimerFunc(REFRESH_MS, onTimer, 0);

  glutMainLoop();
  
  exit();
}
